/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000 Sakai Hiroaki.                              */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Stage 饹(ѤΥ饹)Υ饤֥                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "StageP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

Map StageGetMap(Stage s) {return (s->map);}
ObjList StageGetPlayerList(   Stage s) {return (s->player_list   );}
ObjList StageGetTankList(     Stage s) {return (s->tank_list     );}
ObjList StageGetCanonList(    Stage s) {return (s->canon_list    );}
ObjList StageGetMissileList(  Stage s) {return (s->missile_list  );}
ObjList StageGetExplosionList(Stage s) {return (s->explosion_list);}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

int StageCreateTankForPlayer(Stage stage, Player player)
{
  Tank tank;
  int x, y, tx, ty, f;
  ObjListData current;
  double rx, ry, dx, dy, r;
  Tank other;

  r = ((double)TANK_SIZE / CHARACTER_SIZE) * 2.0;

  /* ᤯̵̤֤ɤå */
  do {
    rx = GetRandomValue(stage->map_width ) + 0.5;
    ry = GetRandomValue(stage->map_height) + 0.5;
    f = 0;
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      other = (Tank)ObjListData_GetObj(current);
      dx = TankGetX(other) - rx;
      dy = TankGetY(other) - ry;
      if (dx * dx + dy * dy <= r * r) {
	f = 1;
	break;
      }
    }
  } while (f);

  tank = TankCreate(rx, ry, GetRandomValue(TANK_ANGLE_NUMBER),
		    player, TANK_ACTIVE, stage->backward);
  if (tank == NULL) ErrorExit("Cannot create tank.");
  PlayerSetTank(player, tank);
  ObjList_InsertObjToEnd(stage->tank_list, tank, (ObjDestructor)TankDestroy);

  /* ֤μ򥯥ꥢ */
  for (x = -1; x <= 1; x++) {
    for (y = -1; y <= 1; y++) {
      tx = ConvertXFromPositionToMap(TankGetX(tank)) + x;
      ty = ConvertYFromPositionToMap(TankGetY(tank)) + y;
      if (MapIsInside(stage->map, tx, ty))
	MapChangeState(stage->map, tx, ty, 0);
    }
  }
  return (0);
}

Stage StageCreate(int stage_number, ObjList player_list,
		  int map_width, int map_height, int width, int height,
		  int peeping, int backward)
{
  Stage stage;
  Player player;
  ObjListData current;

  stage = (Stage)malloc(sizeof(_Stage));
  if (stage == NULL) return (NULL);

  stage->map_width  = map_width;
  stage->map_height = map_height;
  stage->width      = width;
  stage->height     = height;
  stage->peeping    = peeping;
  stage->backward   = backward;

  stage->map = MapCreate(stage, stage->map_width, stage->map_height,
			 stage_number);
  if (stage->map == NULL) ErrorExit("Cannot create stage->map.");

  stage->stage_number = stage_number;
  stage->player_list = player_list;
  stage->tank_list      = ObjList_Create();
  stage->canon_list     = ObjList_Create();
  stage->missile_list   = ObjList_Create();
  stage->explosion_list = ObjList_Create();

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    StageCreateTankForPlayer(stage, player);
  }

  return (stage);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Stage StageDestroy(Stage stage)
{
  ObjListData current;
  Player player;

  if (stage == NULL) return (NULL);

  if (stage->map) MapDestroy(stage->map);

  for (current = ObjList_GetStart(stage->player_list);
       !ObjList_IsEndEdge(stage->player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    PlayerSetTank(player, NULL);
  }
  if (stage->tank_list)      ObjList_Destroy(stage->tank_list);
  if (stage->canon_list)     ObjList_Destroy(stage->canon_list);
  if (stage->missile_list)   ObjList_Destroy(stage->missile_list);
  if (stage->explosion_list) ObjList_Destroy(stage->explosion_list);

  free(stage);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Stage is over?                                                            */
/*---------------------------------------------------------------------------*/

static int StageIsOver(Stage stage, int lose)
{
  switch (ObjList_GetLength(stage->player_list)) {
  case 1 :
    return (lose >= 1);
  case 2 :
    return (lose >= 1);
  case 3 :
    return (lose >= 2);
  case 4 :
    return (lose >= 2);
  case 5 :
    return (lose >= 3);
  default :
    return (lose >= ObjList_GetLength(stage->player_list) * 2 / 3);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ƥȤɽ                                                            */
/*---------------------------------------------------------------------------*/

static int StageSendTextData(Stage stage, Path path)
{
  char buffer[100];
  Player player;
  ObjListData current;

  PathSendNoneData(path, TEXT_CLEAR);
  PathSendShortStringData(path, TEXT_DRAW, 0,
			  "      N.A.D.A.R.\n  Copyright (c) 2000\n    Sakai Hiroaki.\n\n");

  sprintf(buffer, "   Stage : %d\n\n", stage->stage_number + 1);
  PathSendShortStringData(path, TEXT_DRAW, 0, buffer);

  for (current = ObjList_GetStart(stage->player_list);
       !ObjList_IsEndEdge(stage->player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);

    sprintf(buffer, "   %s\n   Win:%d Lose:%d\n\n",
	    PlayerGetName(player),
	    PlayerGetWin(player),
	    PlayerGetLose(player));
    PathSendShortStringData(path, TEXT_DRAW, PlayerGetColorNumber(player),
			    buffer);
  }

  return (0);
}

/*===========================================================================*/
/* ơγ(ФΥᥤ롼)                                      */
/*===========================================================================*/

int StageStartAsServer(Stage stage, ObjList player_list, Path request)
{
  unsigned short int command;
  unsigned short int length;
  int ret;
  char player_name[100];
  int state;
  Path path;
  Player player;
  Player player2;
  Tank tank;
  Canon canon;
  Missile missile;
  Explosion explosion;
  int lose = 0;
  int f0, f1;
  ObjListData current, current2;
  unsigned long int l1;

  /*
   * ХåեΥꥢ
   * Хåե򥯥ꥢƤʤȡơ׵᤬ĤäƤޤ
   * Ѥ߽ŤʤäƤäƤޤǽ롥
   * (SEND_KEY Ф NOW_KEY Υ꥿󤬤ޤäƤޤ̥ꥢ
   *  ӤȿߤʤäƤޤäƤ)
   */

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    while (PathCheckHeader(PlayerGetPath(player), &command, &length)) {

      switch (command) {

      case DELETE_ME :

	PathRecvNoneData(PlayerGetPath(player), &command);
	PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	PlayerSetState(player, PLAYER_DESTROY);
	break;

      default :
	/*
	 * ХåեˤޤäƤ׵򥯥ꥢ롥
	 * (ơ׵᤬ĤäƤޤɤɤѤ߽ŤʤäƤȤ)
	 */
	PathDiscardData(PlayerGetPath(player), &command);
	break;
      }
    }
  }

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    /* ǥץ쥤Υꥢ */
    PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

    /* ɽƥȥǡ */
    StageSendTextData(stage, PlayerGetPath(player));
  }

  /* ޥåץǡ */
  MapSendChangesToAllPlayers(stage->map, stage->player_list);

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    PathSendNoneData(PlayerGetPath(player), READ_KEY);
    PathSendNoneData(PlayerGetPath(player), SEND_KEY);

    /* Хåեեå夷ºݤ˥ǡ */
    PathFlushSendBuffer(PlayerGetPath(player));
  }

  while (1) { /* ơΥᥤ롼 */

    /* 饤Ȥμդ */

    if (PathCheckRecvState(request)) {

      path = PathCreateToCommunicateWithClient(request,
					       SERVER_SEND_BUFFER_SIZE,
					       SERVER_RECV_BUFFER_SIZE);

      while (!PathCheckRecvState(path)) { usleep(10000); }
      ret = PathCheckHeader(path, &command, &length);

      if (ret && (command == JOIN_REQUEST)) {

	PathRecvStringData(path, &command, player_name, 100);

        player = PlayerCreate(player_name, path, PLAYER_ACTIVE);
        ObjList_InsertObjToEnd(player_list, player,
			       (ObjDestructor)PlayerDestroy);

	PathSendShortShortShortShortShortData(PlayerGetPath(player),
					      JOIN_ACCEPT,
					      stage->map_width,
					      stage->map_height,
					      stage->width,
					      stage->height,
					      stage->peeping);
	PathFlushSendBuffer(PlayerGetPath(player));

	PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

	MapSendAllMap(stage->map, player);

	StageCreateTankForPlayer(stage, player);

	/* ɽƥȥǡץ쥤䡼 */
	for (current = ObjList_GetStart(stage->player_list);
	     !ObjList_IsEndEdge(stage->player_list, current);
	     current = ObjListData_GetNext(current)) {
	  player2 = (Player)ObjListData_GetObj(current);
	  if (PlayerGetState(player2) != PLAYER_ACTIVE) continue;
	  StageSendTextData(stage, PlayerGetPath(player2));
	  PathFlushSendBuffer(PlayerGetPath(player2));
	}

	PathSendNoneData(PlayerGetPath(player), READ_KEY);
	PathSendNoneData(PlayerGetPath(player), SEND_KEY);

	/* Хåեեå夷ºݤ˥ǡ */
	PathFlushSendBuffer(PlayerGetPath(player));
      } else {
        path = PathDestroy(path);
      }
    }

    /* ȯν */
    for (current = ObjList_GetStart(stage->explosion_list);
	 !ObjList_IsEndEdge(stage->explosion_list, current);
	 current = ObjListData_GetNext(current)) {
      explosion = (Explosion)ObjListData_GetObj(current);
      state = ExplosionAct(explosion, stage->map);
      switch (state) {
      case EXPLOSION_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->explosion_list, current);
	break;
      default :
	break;
      }
    }

    /* ˤƤν */
    for (current = ObjList_GetStart(stage->canon_list);
	 !ObjList_IsEndEdge(stage->canon_list, current);
	 current = ObjListData_GetNext(current)) {
      canon = (Canon)ObjListData_GetObj(current);
      state = CanonAct(canon, stage);
      switch (state) {
      case CANON_EXPLOSION :
	explosion = ExplosionCreate(CanonGetX(canon), CanonGetY(canon),
				    CanonGetTank(canon), EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	CanonSetState(canon, CANON_DESTROY);
	/* Fall through to destroy the object */
      case CANON_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->canon_list, current);
	break;
      default :
	break;
      }
    }

    /* ߥν */
    for (current = ObjList_GetStart(stage->missile_list);
	 !ObjList_IsEndEdge(stage->missile_list, current);
	 current = ObjListData_GetNext(current)) {
      missile = (Missile)ObjListData_GetObj(current);
      state = MissileAct(missile, stage);
      switch (state) {
      case MISSILE_EXPLOSION :
	explosion = ExplosionCreate(MissileGetX(missile),
				    MissileGetY(missile),
				    MissileGetTank(missile),
				    EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	MissileSetState(missile, MISSILE_DESTROY);
	/* Fall through to destroy the object */
      case MISSILE_DESTROY :
	current = ObjList_DeleteObjToPrev(stage->missile_list, current);
	break;
      default :
	break;
      }
    }

    /*  Player ǡȤ */
    f1 = 1;
    do {
      f0 = 0;
      if (!f1) usleep(1000); /* Wait in 1 millisecond. */
      f1 = 0;

      for (current = ObjList_GetStart(stage->player_list);
	   !ObjList_IsEndEdge(stage->player_list, current);
	   current = ObjListData_GetNext(current)) {

	player = (Player)ObjListData_GetObj(current);
	if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

	f0 = 1;

	if (!PathCheckHeader(PlayerGetPath(player), &command, &length))
	  continue;
	f1 = 1;

	switch (command) {

	case DELETE_ME :

	  PathRecvNoneData(PlayerGetPath(player), &command);
	  PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	  PathFlushSendBuffer(PlayerGetPath(player));
	  TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	  PlayerSetState(player, PLAYER_DESTROY);
	  break;

	case NOW_KEY :

	  PathRecvLongData(PlayerGetPath(player), &command, &l1);
	  TankSetAction(PlayerGetTank(player), l1);
	  PlayerSetState(player, PLAYER_DONE);
	  break;

	default :
	  ErrorExit("unknown command.");
	}
      }
    } while (f0);

    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) == PLAYER_DONE) 
	PlayerSetState(player, PLAYER_ACTIVE);
    }

    /* ֤ν */
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      state = TankAct(tank, stage);
      switch (state) {
      case TANK_EXPLOSION :
	explosion = ExplosionCreate(TankGetX(tank), TankGetY(tank),
				    tank, EXPLOSION_ACTIVE);
	ObjList_InsertObjToEnd(stage->explosion_list, explosion,
			       (ObjDestructor)ExplosionDestroy);
	TankSetState(tank, TANK_EXPLOSING);
	TankSetDestroyTimer(tank, TANK_DESTROY_INTERVAL);
	break;
      case TANK_EXPLOSING :
	TankDecDestroyTimer(tank);
	if (TankGetDestroyTimer(tank) <= 0) {
	  TankSetState(tank, TANK_DESTROY);
	  lose++;
	}
	break;
      case TANK_DESTROY :
	/* current = ObjList_DeleteObjToPrev(stage->tank_list, current); */
	break;
      default :
	break;
      }
    }

    /*
     * ơλȽ
     * (SEND_KEY ׵Ф NOW_KEY ֿƤ饹ơλ
     *  ʤȡNOW_KEY ХåեѤ߽ŤʤäƤơʤ
     *  ȿߤʤȤݤƤޤΤǡNOW_KEY ν
     *  Ȥ˥ơλȽƤ)
     */
    if (StageIsOver(stage, lose)) break;

    /*
     * ǡ
     */

    MapSendChangesToAllPlayers(stage->map, stage->player_list);

    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {

      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

      for (current2 = ObjList_GetStart(stage->tank_list);
	   !ObjList_IsEndEdge(stage->tank_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	tank = (Tank)ObjListData_GetObj(current2);
	/* if (TankGetState(canon) != TANK_ACTIVE) continue; */ /* ס */
	PathSendShortShortShortShortData(PlayerGetPath(player), TANK_DRAW,
					 TankGetX(tank) * 100 + 0.5,
					 TankGetY(tank) * 100 + 0.5,
					 TankGetAngle(tank),
					 TankGetColorNumber(tank));
      }

      for (current2 = ObjList_GetStart(stage->canon_list);
	   !ObjList_IsEndEdge(stage->canon_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	canon = (Canon)ObjListData_GetObj(current2);
	if (CanonGetState(canon) != CANON_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), CANON_DRAW,
			       CanonGetX(canon) * 100 + 0.5,
			       CanonGetY(canon) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->missile_list);
	   !ObjList_IsEndEdge(stage->missile_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	missile = (Missile)ObjListData_GetObj(current2);
	if (MissileGetState(missile) != MISSILE_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), MISSILE_DRAW,
			       MissileGetX(missile) * 100 + 0.5,
			       MissileGetY(missile) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->explosion_list);
	   !ObjList_IsEndEdge(stage->explosion_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	explosion = (Explosion)ObjListData_GetObj(current2);
	if (ExplosionGetState(explosion) != EXPLOSION_ACTIVE) continue;
	PathSendShortShortShortData(PlayerGetPath(player), EXPLOSION_DRAW,
				    ExplosionGetX(explosion) * 100 + 0.5,
				    ExplosionGetY(explosion) * 100 + 0.5,
				    ExplosionGetRadius(explosion));
      }

      PathSendShortShortData(PlayerGetPath(player), DISPLAY_COPY,
			     TankGetX(PlayerGetTank(player)) * 100 + 0.5,
			     TankGetY(PlayerGetTank(player)) * 100 + 0.5);

      PathSendNoneData(PlayerGetPath(player), GAME_CLEAN);
      PathSendNoneData(PlayerGetPath(player), READ_KEY);
      PathSendNoneData(PlayerGetPath(player), SEND_KEY);

      PathFlushSendBuffer(PlayerGetPath(player));
    }

  }

  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
